/*
 *	Copyright (C) 2007, 2008, 2009, 2010. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/// @file
/// @brief File from the QVision library.
/// @author PARP Research Group. University of Murcia, Spain.

#ifndef QVIMAGECANVAS_H
#define QVIMAGECANVAS_H

#include <QPointF>
#include <QVProcessingBlock>
#include <QVPolyline>
#include <QVPolylineF>
#include "qvcanvas.h"

/*!
@class QVImageCanvas qvblockprogramming/qvguiblocks/qvimagecanvas.h QVImageCanvas
@brief Displayer widget for @ref QVImage objects and other data types.

Class @ref QVImageCanvas can be used to create a very flexible and useful widget to show @ref QVImage objects. This widget is in fact a property container, and as such it will read the image to be displayed from an output dynamic property contained in another property container (usually a @ref QVProcessingBlock object).

It is used as follows. First, the programmer must create the QVImageCanvas object in the <i>main()</i> function of a program; then, a link to a QVImage type property from a property holder must be created. For example:

\code
[...]

CannyBlock cannyBlock("Canny");

[...]

QVImageCanvas imageCanvas;
cannyBlock.linkProperty(imageCanvas,"Canny image");

[...]
\endcode

When the application executes, it will automatically create a window like the following:

\image html qvimagecanvas_penguin.png

@note The nice horizontal and vertical scale rules are QWT widgets. The QWT library must be installed, and the QVision must be compiled to use it, for the @ref QVImageCanvas widget to display them. See section @ref OptionalLibraries for more info about this.

You can see it has a zoom number indicator (<i>z = 1</i>), horizontal and vertical rules (in pixels), and some buttons. These can be used to control zooming and moving around a zoomed area of the image. A detailed explanation of each one follows:

<table>
<tr><th><b>Zoom in button</b></th>
\image html qvimagecanvas_zoomin.png
<td>This button zooms the image in. Canvas window doesn't change size, but the area of the image displayed becomes smaller by a factor of 4 (width and height gets divided by 2). The label of the left bottom corner displays the text
\code
z=<zoom>
\endcode
where <i>&lt;zoom&gt;</i> is a number indicating the zoom factor that divides width and height in the actual zoom level.</td></tr>

<tr><th><b>Zoom out button</b></th>
\image html qvimagecanvas_zoomout.png
<td>This button divides zoom factor by 2, if it is equal or greater than 2. If the canvas window is bigger than the image at the final zoom, it is adjusted to the size of the latter.</td></tr>

<tr><th><b>Zoom restore button</b></th>
\image html qvimagecanvas_zoomrestore.png
<td>This button sets zoom factor by 1, and resizes canvas window to its original size.</td></tr>

<tr><th><b>Select zoom region button</b></th>
\image html qvimagecanvas_zoomregion.png
<td>This button lets you select a rectangle in the image to zoom, adjusting zoom factor and canvas window size to it.</td></tr>

<tr><th><b>Select polyline button</b></th>
\image html qvimagecanvas_polyline.png
<td>This button lets you select points in the image. The created points list can then be accessed to in the program by reading the "poly select" imageCanvas output property. You can expand this button by pushing it for a few seconds, and then you will be able to select a polyline representation, a point list representation, or generate a circular polyline, by clicking and dragging.</td></tr>

<tr><th><b>Select ROI button</b></th>
\image html qvimagecanvas_roi.png
<td>This button lets you select a rectangle in the image. That rectangle can be again accessed to using the "rect select" imageCanvas output property. </td></tr>

<tr><th><b>Drag mode</b></th>
\image html qvimagecanvas_drag.png
<td>This button activates drag mode. With it you can move around the image displayed in the canvas window, if the zoom factor forces the canvas window to show only a sub-region of the image, by holding click and dragging on the shown image.</td></tr>
</table>

In the following figure you can see a zoomed canvas window showing a sub-region of the original image:

\image html qvimagecanvas_penguinpixels.png

Another interesting feature of this widget is that at a zoom factor equal or larger than 32, the canvas renders the gray-scale pixel value over every pixel if the image is gray-scale, or the three different values for each of the RGB channels over every pixel, if the image is RGB, as shown below:

\image html qvimagecanvas_penguinpixelnumbers.png

The canvas also knows how to show numeric float values, when showing images of type QVImage<sFloat,1> or QVImage<sFloat,3>.

@ingroup qvblockprogramming
*/
class QVImageCanvas: public QVCanvas, public QVPropertyContainer
	{
		Q_OBJECT
	public:
		/// @brief Image canvas constructor
		///
		/// @param name Name of the image canvas
		/// @param parent parent QObject of the canvas
		QVImageCanvas(const QString name = QString(), QWidget *parent=0);

		/// @brief Sets the display color for a registered input property of the canvas
		///
		/// Registered properties of types QList<QPoint>, QList<QPointF>, QList<QVPolyline> and QList<QVPolylineF> are displayed in the canvas
		/// with a default color assigned by the link method. The user can specify a non-default color value using method @ref setColor.
		///
		/// The following code registers an image property, and a QList<QPointF> property in an image canvas from a source block, and sets the
		/// display color of the points received from the block, through the <i>Detected points</i> property to be red.
		///
		/// @code
		/// QVImageCanvas imageCanvas("Harris corners");
		/// harrisCornersDetector.linkProperty("Input image", imageCanvas);
 		/// // Property "Detected points" is of type QList<QpointF>
		/// harrisCornersDetector.linkProperty("Detected points", imageCanvas);
		/// imageCanvas.setColor("Detected points", Qt::red);
		/// @endcode
		///
		/// @param name Name of the source property to set the color.
		/// @param QColor Color to display the elements contained in the property.
		bool setColor(const QString &name, const QColor &color)
			{ return setPropertyValue<QColor>("Color for " + name, color); }

		/// @todo document
		bool setPrintTags(const QString &name, const bool &printTags)
			{ return setPropertyValue<bool>("Print tags for " + name, printTags); }

		/// @brief Sets the display radius for a registered input property of the canvas
		///
		/// Some registered properties, for example points, are displayed in the canvas with a default radius size of 3. 
		/// As the @ref setColor, this method can be used to specify a display radius size different than the default.
		///
		/// @param name Name of the source property to set the radius.
		/// @param radius Radius size of the property.
		bool setRadius(const QString &name, const int &radius)
			{ return setPropertyValue<int>("Radius for " + name, radius); }

		/// @brief Sets the saturation range to display floating point images
		///
		/// The default saturation range to display a floating point image (QVImage<sFloat, 1> or QVImage<sFloat, 3>) is [0, 255]. This means that
		/// pixel intensity values outside that range will be set to 0 or 255, wether the intensity is below or above that range.
		/// 
		/// This method can be used to specify a different saturation range.
		///
		/// @param low Lower value for the saturation range.
		/// @param high Higher value for the saturation range.
		void setSaturationRange(const float low, const float high);

		/// @brief Link selected polyline property with a block's input property.
		/// @todo expand this documentation
		bool linkSelectedPolyline(QVPropertyContainer *destinationContainer, QString destinationPropName)
			{
			return linkProperty("poly select", destinationContainer, destinationPropName);
			}

		/// @brief Link selected rectangle property with a block's input property.
		/// @todo expand this documentation
		bool linkSelectedRectangle(QVPropertyContainer *destinationContainer, QString destinationPropName)
			{
			return linkProperty("rect select", destinationContainer, destinationPropName);
			}

		/// @brief Link selected polyline property with a block's input property.
		///
		/// This is an overloaded function provided for convenience.
		/// @see linkSelectedPolyline(QVPropertyContainer *, QString)
		bool linkSelectedPolyline(QVPropertyContainer &destinationContainer, QString destinationPropName)
			{
			return linkSelectedPolyline(&destinationContainer, destinationPropName);
			}

		/// @brief Link selected rectangle property with a block's input property.
		///
		/// This is an overloaded function provided for convenience.
		/// @see linkSelectedRectangle(QVPropertyContainer *, QString)
		bool linkSelectedRectangle(QVPropertyContainer &destinationContainer, QString destinationPropName)
			{
			return linkSelectedRectangle(&destinationContainer, destinationPropName);
			}

		/// @brief Display user defined data types method
		///
		/// By reimplementing this method on @ref QVImageCanvas subclasses, the developer can modify the way the objects are displayed in the canvas, or
		/// specify the way new user data type objects must be displayed.
		/// 
		/// These subclasses can create image canvas objects which display registered objects the new way.
		/// 
		/// Method @ref custom_viewer is called whenever the canvas must be repainted, for example due to changes in the input properties or a
		/// canvas resizement. The reimplemented code for this method must read the user defined properties, and use @ref QVImageCanvas or QWidget
		/// functionality to correctly display them.
		///
		/// An example @ref custom_viewer function reimplementation follows:
		///
		/// @code
		/// void MyImageCanvas::custom_viewer()
		/// 	{
		///	foreach(QString propertyName, getPropertyListByType< QList< QVPolylineF > >())
		///		{
		///		QList< QVPolylineF > polylineList = getPropertyValue< QList< QVPolylineF > >(propertyName);
		///		foreach(QVPolylineF polyline, polylineList)
		///			{
		///			// Set the color of the polyline
		///			const int pointHash = qHash(polyline.first());
		///			draw(polyline, qvColors[pointHash % 10]);
		///
		///			// Draw red dots on the end of the polyline
		///			getQVPainter()->setPen(Qt::red);
		///			getQVPainter()->setBrush(QBrush(Qt::red, Qt::SolidPattern));
		///			getQVPainter()->drawEllipse(polyline.first().x()-1, polyline.first().y()-1, 3, 3);
		///			getQVPainter()->drawEllipse(polyline.last().x()-1, polyline.last().y()-1, 3, 3);
		///			}
		///		}
		///	}
		/// @endcode
		///
		/// Image canvas objects created from the class <i>MyImageCanvas</i>, which supossely inherits from @ref QVImageCanvas will display polylines
		/// contained in input polyline lists with different colors, depending on the location of their first point, and red dots of radius 3 at the
		/// ends of each polyline.
		virtual void custom_viewer() { };

		/// @deprecated use @ref QVImageCanvas::setSaturationRange(const float,const float) instead.
		void setLowHigh(float low,float high)
			{
			std::cout << "DEPRECATED, use setSaturationRange instead" << std::endl;
			setSaturationRange(low, high);
			}

		void unlink();

		#ifndef DOXYGEN_IGNORE_THIS
		void viewer();
		#endif

	public slots:
		void rectSelectedSlot(QRect rect);
		void polySelectedSlot(QPoint point, bool reset, TPolyMode mode);
		void circleSelectedSlot(QPoint center, float radius);

	protected:
		bool linkUnspecifiedInputProperty(QVPropertyContainer *sourceContainer, QString sourcePropName, LinkType linkType = AsynchronousLink);
		bool linkUnspecifiedOutputProperty(QVPropertyContainer *destContainer, QString destPropName, LinkType linkType = AsynchronousLink);
		bool treatUnlinkInputProperty(QString destPropName, QVPropertyContainer *sourceCont, QString sourcePropName);

		void draw(const QList<QPoint> &pointList, QColor color = Qt::red, bool printTags = false, int radius = 3);
		void draw(const QList<QPointF> &pointList, QColor color = Qt::red, bool printTags = false, double radius = 3);
		void draw(const QVPolyline &polyline, QColor color = Qt::red, bool printTags = false);
		void draw(const QVPolylineF &polylinef, QColor color = Qt::red, bool printTags = false);
		void draw(const QRect &rectangle, QColor color = Qt::red, bool printTags = false);

		void closeEvent(QCloseEvent *event) { Q_UNUSED(event); emit closed(); }
	private:

		const QColor	getNextColor()
			{
			QColor color = qvColors[colorCursor++];
			colorCursor %= 10;

			return color;
			}

		float _low,_high;
		int colorCursor, contentLinkedBlocks;

	signals:
		void closed();

	};

Q_DECLARE_METATYPE(TPolyMode);

#endif
