/*
 *	Copyright (C) 2007, 2008, 2009, 2010. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/// @file
/// @brief File from the QVision library.
/// @author PARP Research Group. University of Murcia, Spain.

#ifndef QMATRIXALGEBRA_H
#define QMATRIXALGEBRA_H

#include <QVector>
#include <QVMatrix>

/// @brief Obtains the <a href="http://en.wikipedia.org/wiki/Singular_value_decomposition">singular value decomposition (SVD)</a> for a matrix.
///
/// Resulting matrices <i>U</i>, <i>V</i> and <i>S</i> will satisfy the following relationship with the input matrix:
///
/// \f$ M = U S V^T \f$
///
/// @ingroup qvmatrixalgebra
/// @param M matrix containing matrix to decompose
/// @param U matrix to store resulting matrix U from decomposition
/// @param V matrix to store resulting matrix V from decomposition
/// @param S matrix to store resulting matrix S from decomposition
/// @deprecated use @ref SingularValueDecomposition(const QVMatrix &, QVMatrix &, QVMatrix &, QVMatrix &) instead.
void singularValueDecomposition(const QVMatrix &M, QVMatrix &U, QVMatrix &V, QVMatrix &S);

/*! @brief Obtains the <a href="http://en.wikipedia.org/wiki/Singular_value_decomposition">Singular Value Decomposition (SVD)</a> of a matrix.

The SV decomposition obtains three matrices: <i>U</i> <i>S</i>, and <i>V</i> from an original matrix
<i>A</i> satisfying the following equation:

\f$ A = U S V^T \f$

The matrix <i>U</i> is an upper triangular matrix. The matrix <i>S</i> is a diagonal matrix, and the matrix <i>V</i> is an orthogonal matrix
(that is, \f$ V V^T = I \f$).

@param A matrix to decompose
@param U param to store the matrix U from the SVD decomposition of A
@param S param to store the matrix S from the SVD decomposition of A
@param V param to store the matrix V from the SVD decomposition of A
@ingroup qvmatrixalgebra
*/
void SingularValueDecomposition(const QVMatrix &A, QVMatrix &U, QVMatrix &S, QVMatrix &V);

/*!
@brief Obtains the <a href="http://en.wikipedia.org/wiki/LU_decomposition">LU decomposition</a> for a matrix.

The Cholesky decomposition obtains three matrices: <i>L</i> <i>U</i>, and <i>P</i> from an original matrix
<i>A</i>, satisfying the following equation:

\f$ A = P L U \f$

Matrices <i>L</i> and <i>U</i> are a lower and upper triangular matrices respectively. <i>P</i> is a permutation
matrix, which is equal to the identity matrix if <i>A</i> is invertible.

@param A param containing the matrix to decompose
@param L param to store the resulting matrix L from the LU decomposition of A
@param U param to store the resulting matrix U from the LU decomposition of A
@param P param to store the resulting matrix P from the LU decomposition of A
@ingroup qvmatrixalgebra
*/
void LUDecomposition(const QVMatrix &A, QVMatrix &L, QVMatrix &U, QVMatrix &P);

/*!
@brief Obtains the <a href="http://en.wikipedia.org/wiki/Cholesky_decomposition">Cholesky decomposition</a> for a matrix.

The Cholesky decomposition obtains a matrix <i>L</i> from an original matrix <i>A</i>, satisfying the following equation:

\f$ A = L L^{*} \f$

The matrix \f$ L \f$ is a lower triangular matrix, with strictly positive diagonal entries. The \f$ L^{*} \f$ is the <a href="http://en.wikipedia.org/wiki/Conjugate_transpose">conjugate transpose</a> of the matrix \f$ L \f$.

@param A param containing the matrix to decompose
@param L param to store the resulting matrix L from the Cholesky decomposition of A
@ingroup qvmatrixalgebra
*/
void CholeskyDecomposition(const QVMatrix &A, QVMatrix &L);

/*! @brief Obtains the <a href="http://en.wikipedia.org/wiki/QR_decomposition">QR decomposition</a> for a matrix.

The QR decomposition obtains two matrices <i>Q</i> and <i>R</i> from an original matrix <i>A</i> that satisfy the following equation:

\f$ A = Q R \f$

Matrix <i>Q</i> is orthogonal (that is, \f$ Q Q^T = I \f$) and <i>R</i> is an upper triangular matrix.

@ingroup qvmatrixalgebra
@param A param containing the matrix to decompose
@param Q param to store the resulting matrix Q from the QR decomposition of A
@param R param to store the resulting matrix R from the QR decomposition of A
*/
void QRDecomposition(const QVMatrix &A, QVMatrix &Q, QVMatrix &R);

/*! @brief Obtains the <a href="http://en.wikipedia.org/wiki/Moore-Penrose_pseudoinverse">Moore–Penrose pseudoinverse</a> for a matrix.

The Moore–Penrose pseudoinverse of an initial matrix <i>A</i> is another matrix <i>A<sup>+</sup></i> that satisfies the following equation:

\f$ A A^+ = I \f$

For the matrix <i>A<sup>+</sup></i> to exist, these conditions must apply to matrix <i>A</i>:
	- Its first dimension must be lower or equal than its second dimension.
	- Its determinant must not be zero.

@param A matrix containing matrix to obtain pseudoinverse
@ingroup qvmatrixalgebra
*/
QVMatrix pseudoInverse(const QVMatrix &A);

/*! @brief Obtains the <a href="http://en.wikipedia.org/wiki/Eigendecomposition_of_a_matrix">eigen-decomposition</a> for a symetric matrix.

The eigen-decomposition obtains the factorization of an initial symetric matrix <i>A</i> into a canonical form, represented
in terms of a set of (eigen) vectors, and their corresponding (eigen) values.

Each eigenvector \f$ \mathbf{v} \f$ and its corresponding eigenvalue \f$ \lambda \f$ satisfy the following equation.

\f$ A \mathbf{v} = \lambda \mathbf{v} \f$

This function returns the eigenvectors and their corresponding eigenvalues in a <i>Q</i> matrix and a <i>l</i> vector.
Each eigenvector is stored as a row of the <i>Q</i> matrix. The <i>i</i>-th element of the vector <i>l</i> contains the eigenvalue
corresponding to the eigenvector stored in the <i>i</i>-th row.

The following equation holds, given a matrix \f$ Q \f$ and the vector \f$ \vec{\lambda} \f$ containing respectively the eigen-vectors and the eigen-values of the symetric matrix <i>A</i>:

\f$ A = Q^{T}diag(\vec{\lambda})Q^{-T} \f$ 

@param A symetric matrix to obtain eigen-decomposition
@param l vector containing the eigenvalues from the eigen-decomposition of A
@param Q matrix containing the eigenvectors from the eigen-decomposition of A. Eigenvectors are stored as row vectors.
@ingroup qvmatrixalgebra
*/
void eigenDecomposition(const QVMatrix &A, QVVector &l, QVMatrix &Q);

/*! @brief Obtains the <a href="http://en.wikipedia.org/wiki/Determinant">determinant</a> of a squared matrix.

@param A matrix to obtain the determinant
@return the value of the determinant
@ingroup qvmatrixalgebra
*/
double determinant(const QVMatrix &A);

/*! @brief Solves a system of linear equations using a <a href="http://en.wikipedia.org/wiki/Householder_transformation">Householder transformation</a>.

Given a coefficient matrix <i>A</i> and an objective vector <i>b</i>, this functions obtain the vector <i>x</i> satisfying the following equation:

\f$ A\mathbf{x} = \mathbf{b} \f$

For a more efficient way of solving several systems of linear equations simultaneously, see
@ref solveLinear(const QVMatrix &, QVMatrix &, const QVMatrix &) and @ref solveOverDetermined(const QVMatrix &, QVMatrix &, const QVMatrix &) functions.

@param A coeficient matrix in the matrix equation form of the problem.
@param x vector to store the solution.
@param b right-hand side vector in the matrix equation form of the problem.
@see solveLinear(const QVMatrix &, QVVector &, const QVVector &)
@see solveLinear(const QVMatrix &, QVMatrix &, const QVMatrix &)
@see solveOverDetermined(const QVMatrix &, QVMatrix &, const QVMatrix &)
@ingroup qvmatrixalgebra
*/
void solveLinear(const QVMatrix &A, QVVector &x, const QVVector &b);

/*! @brief Solves several system of linear equations, using a LU decomposition.

Given a coefficient matrix <i>A</i> and an objectives matrix <i>B</i>, this functions obtain the matrix <i>x</i> satisfying the following equation:

\f$ AX = B \f$

Notice that the solution of this equation is equivalent to solving the equations:

\f$ A\mathbf{x_i} = \mathbf{b_i} \f$

Where <i>x<sub>i</sub></i> and <i>b<sub>i</sub></i> are respectively the <i>i</i>-th columns of <i>X</i> and <i>B</i>. But obtaining a solution
for the latter set of equations is less efficient and worst conditioned than solving the former matrix equation.

This function is based on a LU decomposition, to obtain the solution for the equation. Thus, the system can not be overdetermined. That is, the
<i>A</i>, <i>X</i> and <i>B</i> must be square.

@param A coeficient matrix in the matrix equation form of the problem.
@param x matrix to store the solutions, each column is a solution.
@param B right-hand side matrix in the matrix equation form of the problem, each column is a right-hand side.
@see solveOverDetermined(const QVMatrix &, QVMatrix &, const QVMatrix &)
@see solveLinear(const QVMatrix &, QVVector &, const QVVector &)
@ingroup qvmatrixalgebra
*/
void solveLinear(const QVMatrix &A, QVMatrix &X, const QVMatrix &B);

/*! @brief Solves several system of linear equations, using a SV decomposition.

Given a coefficient matrix <i>A</i> and an objectives matrix <i>B</i>, this functions obtain the matrix <i>x</i> satisfying the following equation:

\f$ AX = B \f$

This function is equivalent to @ref solveLinear(const QVMatrix &, QVMatrix &, const QVMatrix &). The difference is that it uses SV decomposition,
thus it can handle over-determined systems, where the first dimension of the matrix <i>B<i> is smaller or equal than its second dimension.

@param A coeficient matrix in the matrix equation form of the problem.
@param X matrix to store the solutions, each column is a solution.
@param B right-hand side matrix in the matrix equation form of the problem, each column is a right-hand side.
@see solveLinear(const QVMatrix &, QVMatrix &, const QVMatrix &)
@see solveLinear(const QVMatrix &, QVVector &, const QVVector &)
@ingroup qvmatrixalgebra
*/
void solveOverDetermined(const QVMatrix &A, QVMatrix &X, const QVMatrix &B);

/*! @brief Solves an <a href="http://en.wikipedia.org/wiki/System_of_linear_equations#Homogeneous_systems">homogeneous linear system</a>.

Given a matrix <i>A</i>, this functions obtain the vector <i>x</i> satisfying the following equation:

\f$ A\mathbf{x} = \mathbf{0} \f$

The solution is based on the SV decomposition of the matrix <i>A</i>. Vector <i>x</i> is set to the last column
of the matrix <i>V</i> from the SV decomposition of <i>A</i>.

@param A coeficient matrix for the homogeneous equation system.
@param x vector to store the solution.
@ingroup qvmatrixalgebra
*/
void solveHomogeneousLinear(const QVMatrix &A, QVector<double> &x);

#ifndef DOXYGEN_IGNORE_THIS
/// @todo document this function
void solveHomogeneousLinear2(const QVMatrix &A, QVector<double> &x);

/// @todo document this function
double homogLineFromMoments(double x,double y,double xx,double xy,double yy,double &a,double &b,double &c);

/// @todo document and test this function
QVVector regressionLine(const QVMatrix &points);
#endif

#endif
